/*
=====================================================================
	BITMAP.CPP (Version 1.04)
		Source code to bitmap class (uncompressed .BMP).
	Copyright (c) 2004 Daniel R. Collins. All rights reserved.
	See the bottom of this file for any licensing information.
=====================================================================
*/

#include <stdio.h>
#include <assert.h>
#include "bitmap.h"


/*
---------------------------------------------------------------------
  Definitions
---------------------------------------------------------------------
*/

// Type definitions

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned int DWORD;
typedef signed int LONG;


// File header

typedef struct tagBITMAPFILEHEADER { // bmpfh 
    WORD    bfType;
    DWORD   bfSize;
    WORD    bfReserved1;
    WORD    bfReserved2;
    DWORD   bfOffBits;
} BITMAPFILEHEADER;


// Info header

typedef struct tagBITMAPINFOHEADER { // bmpih 
    DWORD  biSize; 
    LONG   biWidth; 
    LONG   biHeight; 
    WORD   biPlanes; 
    WORD   biBitCount; 
    DWORD  biCompression; 
    DWORD  biSizeImage; 
    LONG   biXPelsPerMeter; 
    LONG   biYPelsPerMeter; 
    DWORD  biClrUsed; 
    DWORD  biClrImportant; 
} BITMAPINFOHEADER; 


// RGB quad

typedef struct tagRGBQUAD { // rgbq 
    BYTE    rgbBlue; 
    BYTE    rgbGreen; 
    BYTE    rgbRed; 
    BYTE    rgbReserved; 
} RGBQUAD; 


// Color map

#define COLORMAP_SIZE 256
typedef RGBQUAD COLORMAP[COLORMAP_SIZE];


/*
---------------------------------------------------------------------
  Internal functions
---------------------------------------------------------------------
*/

// Calculate proper scanline size (multiple of 4 bytes)

int bmpScanLineSize (int width, int depth) {
  int scanLineSize = width * depth;
  int offset = scanLineSize % 4;
  return scanLineSize + (offset ? 4-offset : 0);
}


// Fill a bitmap file header

void bmpFileHeaderFill 
  (BITMAPFILEHEADER *bmpfh, int width, int height, int depth) {

  assert(bmpfh);
  ((char*)(&(bmpfh->bfType)))[0] = 'B';
  ((char*)(&(bmpfh->bfType)))[1] = 'M';
  bmpfh->bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
  bmpfh->bfSize = bmpfh->bfOffBits + width * height * depth;
  bmpfh->bfReserved1 = bmpfh->bfReserved2 = 0;
}


// Fill a bitmap info header

void bmpInfoHeaderFill 
  (BITMAPINFOHEADER *bmpih, int width, int height, int depth) {

  assert(bmpih);
  bmpih->biSize = 40;
  bmpih->biWidth = width;
  bmpih->biHeight = height;
  bmpih->biPlanes = 1;
  bmpih->biBitCount = depth * 8;
  bmpih->biCompression = 0;
  bmpih->biSizeImage = width * height;
  bmpih->biXPelsPerMeter = 3780;
  bmpih->biYPelsPerMeter = 3780;
  bmpih->biClrUsed = 0;
  bmpih->biClrImportant = 0;
}


// Fill a grayscale color map

void bmpColorMapFillGray (COLORMAP cm) {
  for (int i = 0; i < COLORMAP_SIZE; i++) {
    cm[i].rgbBlue = cm[i].rgbGreen = cm[i].rgbRed = i;
    cm[i].rgbReserved = 0;
  }
}


/*
---------------------------------------------------------------------
  External function
---------------------------------------------------------------------
*/

// Save pixel data to disk as .BMP file

void bmpDataSave 
  (void *data, int width, int height, int depth, const char *filename) {

  // Width must be a multiple of 4
  assert(width % 4 == 0);

  // Handle either 1 or 3 bytes per pixel
  assert(depth == BMP_DEPTH_GRAYSCALE || depth == BMP_DEPTH_COLOR);

  // Make sure compiler doesn't alter size of structs
  assert(sizeof(BITMAPFILEHEADER) == 14);

  // Make file header
  BITMAPFILEHEADER bmpfh;
  bmpFileHeaderFill(&bmpfh, width, height, depth);

  // Make info header
  BITMAPINFOHEADER bmpih;
  bmpInfoHeaderFill(&bmpih, width, height, depth);

  // If grayscale, add colormap
  COLORMAP cm;
  if (depth == BMP_DEPTH_GRAYSCALE) {
    bmpColorMapFillGray(cm);    
    bmpfh.bfOffBits += sizeof(COLORMAP);
    bmpfh.bfSize += sizeof(COLORMAP);
  }

  // Save to file
	FILE *f;
	f = fopen(filename, "wb");
  fwrite(&bmpfh, sizeof(BITMAPFILEHEADER), 1, f);
  fwrite(&bmpih, sizeof(BITMAPINFOHEADER), 1, f);
  if (depth == BMP_DEPTH_GRAYSCALE) {
    fwrite(&cm, sizeof(COLORMAP), 1, f);
  }
  fwrite(data, depth, width * height, f);
	fclose(f);
}


/*
=====================================================================
LICENSING INFORMATION

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The author may be contacted by email at: dcollins@superdan.net
=====================================================================
*/

